/*
 *
 * settihtc/gettihtc: 
 *   switch serial line from/to N_TIHTC kine discipline or read
 *   its state. optionally forks into background
 *
 * This is a modified version of test_tihtc.c which is
 * Copyright (C) 2007 Philipp Zabel <philipp.zabel_at_gmail.com> 
 *
 * Modifications: Albrecht Gebhardt <alge@algepop.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
 

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <libgen.h>
#include <sys/types.h>
#include <unistd.h>
#include <dirent.h>
#include <signal.h>
#include <dirent.h>

#ifndef N_TTY
#define N_TTY 0
#endif
#ifndef N_TIHTC
#define N_TIHTC 17
#endif

#define CHAN_TIHTC_ATCMD     0x16
#define CHAN_TIHTC_DEBUG     0x12
#define CHAN_TIHTC_CONFIG    0x19


/* stolen from HCIUART set/get protocol ioctls */
#define TIHTCSETCHAN _IOW('U', 200, int)
#define TIHTCGETCHAN _IOR('U', 201, int)

#define MAXPLEN 128

int main(int argc, char *argv[])
{
    int i, fd, ldisc = N_TIHTC, channel = CHAN_TIHTC_ATCMD;
    int keepalive=0, killold=0, n_lnname;
    char *pname, *devname;
    pid_t pid, oldpid, i_pid;
    FILE *pidfile;
    DIR *proc;
    struct dirent *entry;
    char *name, procname[MAXPLEN], lnname[MAXPLEN];

    pname = strdup(argv[0]);

    if (argc < 2 || argc > 4) {
	if(strncmp(basename(pname),"settihtc",8)==0){
	    printf("Usage: %s <dev> [atcmd|debug|config] [-k]\n", argv[0]);
	    printf("       default channel: atcmd\n");
	    printf("       -k  keepalive, forks child which keeps <dev> opened,\n");
	    printf("reset to non-tihtc line discipline, kills \"%s -k\" if necessary:\n", argv[0]);
	    printf("       %s <dev> -r\n", argv[0]);
	    return -1;
	} else if(strncmp(basename(pname),"gettihtc",8)==0){
	    printf("read line discipline and channel:\n");
	    printf("Usage: %s <dev> \n", argv[0]);
	    return -1;
	}
    }
    devname = strdup(argv[1]);

    if(strncmp(basename(pname),"settihtc",8)==0){
	for(i=2;i<argc;i++){
	    if(strncmp(argv[i],"atcmd",5)==0){
		channel=CHAN_TIHTC_ATCMD;
	    } else if(strncmp(argv[i],"debug",5)==0) {
		channel=CHAN_TIHTC_DEBUG;
	    } else if(strncmp(argv[i],"config",6)==0) {
		channel=CHAN_TIHTC_CONFIG;
	    } else if(strncmp(argv[i],"-r",2)==0) {
		ldisc=N_TTY;
                killold=1;
            } else if(strncmp(argv[i],"-k",2)==0) {
                keepalive=1;
	    } else {
		printf("unknown option: %s\n",argv[i]);
		return -1;
	    }
        }
    }
    fd = open(devname, O_RDWR);
    if (fd < 0) {
	printf("Can't open %s\n", argv[1]);
	return -1;
    }

    if(strncmp(basename(pname),"settihtc",8)==0){
	/* Set line discipline */
//	printf("debug: setting %s to ldisc %d\n", devname, ldisc);
	if (ioctl (fd, TIOCSETD, &ldisc) < 0) {
	    perror("Can't set line discipline");
	    return -1;
	}

	/* Set channel */
        if(ldisc!=N_TTY)
	  if (ioctl(fd, TIHTCSETCHAN, channel) < 0) {
	      perror("Can't set channel");
	      return -1;
  	  }
        if(killold){
          pidfile = fopen("/var/run/settihtc.pid","r");
          if(pidfile){
            if(fscanf(pidfile,"%d",&oldpid)==1){
              fclose(pidfile);
              proc = opendir("/proc");
              if(proc){
                for(;;) {
                  if((entry = readdir(proc)) == NULL) {
                    break;
                    ;
                  }
                  name = entry->d_name;
                  if (!(*name >= '0' && *name <= '9'))
                    continue;
                  i_pid=atoi(name);
                  if(oldpid == i_pid){
                    kill(oldpid, SIGKILL);
                    unlink("/var/run/settihtc.pid");
                    break;
                  }
                }
                closedir(proc);
              }
            } 
          }
        }
        if(ldisc==N_TTY)
            return 0;

        if(keepalive){
          pidfile = fopen("/var/run/settihtc.pid","r");
          if(pidfile){
           if(fscanf(pidfile,"%d",&oldpid)==1){
             snprintf(procname, 15, "/proc/%d/exe",oldpid);
             n_lnname = readlink(procname, lnname, MAXPLEN);
             if(n_lnname>0){
               procname[n_lnname]='\0';
               if(strstr(procname, "settihtc")){
                 printf("removing stale pid file ...\n");
                 fclose(pidfile);
                 unlink("/var/run/settihtc.pid");
               } else {
                 printf("settihtc already running in background with PID %d\n", oldpid);
                 fclose(pidfile);
                 return 1;
               }
             } else {
               printf("removing stale pid file ...\n");
               fclose(pidfile);
               unlink("/var/run/settihtc.pid");
             }
           } else {
             printf("removing stale pid file ...\n");
             fclose(pidfile);
             unlink("/var/run/settihtc.pid");
           };
          }
          pid=fork();
          if(pid!=0){
            printf("settihtc daemon forked with PID %d\n", pid);
	    pidfile = fopen("/var/run/settihtc.pid","w+");
            if(pidfile){
	      fprintf(pidfile,"%d",pid);
              fclose(pidfile);
            }
            return 0;
          } else {
            /* go into indefinite loop: */
            while(1){
              sleep(60);
            };
          }
        }
    } else if(strncmp(basename(pname),"gettihtc",8)==0){
        /* Get line discipline */
        if (ioctl (fd, TIOCGETD, &ldisc) < 0) {
	    perror("Can't get line discipline");
	    return -1;
        }
	
        if(ldisc==N_TTY){
	    printf("line discipline for %s: TTY\n", devname);
	    return 0;
        }

        if(ldisc==N_TIHTC){
	    printf("line discipline for %s: TIHTC\n", devname);
	}

        /* Get channel */
        if (ioctl(fd, TIHTCGETCHAN, &channel) < 0) {
	    perror("Can't get channel");
	    return -1;
        }
	if(channel==CHAN_TIHTC_ATCMD)
	    printf("tihtc channel: atcmd\n");
        if(channel==CHAN_TIHTC_DEBUG)
	    printf("tihtc channel: debug\n");
        if(channel==CHAN_TIHTC_CONFIG)
	    printf("tihtc channel: config\n");

    }  


    close(fd);
    return 0;
      
}

